/*
   Noiasca Smartmeter Library 
   
   copyright 2021 noiasca noiasca@yahoo.com
   
   credits:
   - The Serial Receiver is based on the examples of Serial Input Basics  
   
   by noiasca
   
   Version
   2021-12-12 0.0.1 initial version

*/
#pragma once
#include <Arduino.h>

#define NOIASCA_SMARTMETER_VERSION "Noiasca Smartmeter 0.0.1"   // this library

#define  SMARTMETER_DEBUG  0 // Turn on debug statements to the serial output

#if SMARTMETER_DEBUG
#define SM_DEBUG(...) Serial.print(__VA_ARGS__)
#define SM_DEBUGLN(...) Serial.println(__VA_ARGS__)
#else
#define SM_DEBUG(...)
#define SM_DEBUGLN(...) 
#endif
  
constexpr byte tarifs = 2;                       // how many tarifs are used for "electricityDelivered"

struct SM_Data
{
  char timestamp[14];                            // 13 + null
  char equipmentIdentifier[32];                  // up to 96 in protocol
  float electricityDeliveredToClient[tarifs];    // up to 16 tarifs (see chapter 7 page 26)
  float electricityDeliveredByClient[tarifs];
  float instantaneousVoltageL1;
  float instantaneousVoltageL2;
  float instantaneousVoltageL3;
  float instantaneousCurrentL1;
  float instantaneousCurrentL2;
  float instantaneousCurrentL3;
  float actualElectricityPowerReceived;
  float actualElectricityPowerDelivered;
  bool isDirty = true;                           // at least one value was changed
};

class SmartmeterP1 {
    SM_Data data;                                // stores the last received data
    static const byte numChars = 32;             // receiver buffer size
    char receivedChars[numChars];                // an array to store the received data
    boolean newData = false;                     // flag if receiver has data in buffer
    byte ndx = 0;                                // current position in receive buffer/length
    Stream &stream;                              // reference to a serial interface, can be HW Serial or software serial
    void (*cbOnNewData)();                       // gets called when a full message was processed
    
  public:
    SmartmeterP1 (Stream &stream) : stream(stream) {}
    
    /*
       Several get functions to return the values from data
    */
    char * getTimestamp()    {return data.timestamp;}
    char * getEquipmentIdentifier()    {return data.equipmentIdentifier;}
    float getElectricityDeliveredToClient(byte tarif = 0) 
    { 
      if (tarif < tarifs) 
        return data.electricityDeliveredToClient[tarif];
      else
        return 0;
    }
    float getElectricityDeliveredByClient(byte tarif = 0) 
    {
      if (tarif < tarifs)
        return data.electricityDeliveredToClient[tarif];
      else
        return 0;
    }
    float getInstantaneousVoltageL1()            {return data.instantaneousVoltageL1;}
    float getInstantaneousVoltageL2()            {return data.instantaneousVoltageL2;}
    float getInstantaneousVoltageL3()            {return data.instantaneousVoltageL3;}
    float getInstantaneousCurrentL1()            {return data.instantaneousCurrentL1;}
    float getInstantaneousCurrentL2()            {return data.instantaneousCurrentL2;}
    float getInstantaneousCurrentL3()            {return data.instantaneousCurrentL3;}
    float getActualElectricityPowerReceived()    {return data.actualElectricityPowerReceived;}
    float getActualElectricityPowerDelivered()   {return data.actualElectricityPowerDelivered;}
    bool isDirty()                               {return data.isDirty;}     
    
    /*
       set the callback
    */    
    void setOnNewData(void (*cbOnNewData)())
    {
      (*this).cbOnNewData = cbOnNewData;
    }

    /*
       the "run" method
       must be called in loop()
       it is based on the proposals of "Serial Input Basics" Example 4
       was recvWithEndMarker();
    */      
    void update()
    {
      char endMarker = '\n';
      char rc;
      if (stream.available() > 0) {
        rc = stream.read();
        if (rc != endMarker) {
          receivedChars[ndx] = rc;
          ndx++;
          if (ndx >= numChars) {
            ndx = numChars - 1;
          }
        }
        else {
          receivedChars[ndx] = '\0';   // terminate the string
          ndx = 0;
          newData = true;
          parseLine();
        }
      }
    }

    /*
       interprete one line of received characters
       update the value in data structure
       handles the dirty flag
    */ 
    void parseLine() {
      if (newData == true) {
        // output all
        SM_DEBUG(F("D133 received "));
        SM_DEBUGLN(receivedChars);

        char* ptrBracket = strchr(receivedChars, '(');
        if (!strncmp(receivedChars, "0:1.0.0(", 8))
        {
          //SM_DEBUG(F("D139 time"));
          strncpy(data.timestamp, ptrBracket + 1, sizeof(data.timestamp) - 1);
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "0-0:96.1.1(", 11))
        {
          const size_t length = sizeof(data.equipmentIdentifier) - 1;
          size_t lengthRec = strlen(receivedChars) - 12;
          lengthRec = min(length, lengthRec);
          strncpy(data.equipmentIdentifier, ptrBracket + 1, lengthRec);
          data.equipmentIdentifier[lengthRec] = '\0';
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "1-0:1.7.0(", 10))
        {
          data.actualElectricityPowerDelivered = strtod(ptrBracket + 1, NULL);
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "1-0:2.7.0(", 10))
        {
          data.actualElectricityPowerReceived = strtod(ptrBracket + 1, NULL);
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "1-0:1.8.", 8))   // we might receive this information for several tarifs (up to 16)
        {
          byte tarif = atoi(receivedChars + 8);
          if (tarif <= tarifs && tarif <= 16 && tarif >= 1)
          {
            tarif--;         // array starts counting at 0, but protocol starts with 1
            //SM_DEBUG(F("D168 toclient=")); 
            //SM_DEBUGLN(tarif);
            data.electricityDeliveredToClient[tarif] = strtod(ptrBracket + 1, NULL);
            data.isDirty = true;
          }
          else
          {
            SM_DEBUG(F("E175: invalid Tarif=")); SM_DEBUG(tarif);
          }
        }
        else if (!strncmp(receivedChars, "1-0:2.8.", 8))
        {
          byte tarif = atoi(receivedChars + 8);
          if (tarif <= tarifs && tarif > 0)
          {
            tarif--;         // array starts counting at 0, but protocol starts with 1
            //SM_DEBUG(F("D181 toclient=")); 
            //SM_DEBUG(tarif);
            data.electricityDeliveredByClient[tarif] = strtod(ptrBracket + 1, NULL);
            data.isDirty = true;
          }
          else
          {
            SM_DEBUG(F("E191: invalid Tarif ")); SM_DEBUG(tarif);
          }
        }
        else if (!strncmp(receivedChars, "1-0:32.7.0(", 10))
        {
          data.instantaneousVoltageL1 = strtod(ptrBracket + 1, NULL);
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "1-0:52.7.0(", 10))
        {
          data.instantaneousVoltageL2 = strtod(ptrBracket + 1, NULL);
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "1-0:72.7.0(", 10))
        {
          data.instantaneousVoltageL3 = strtod(ptrBracket + 1, NULL);
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "1-0:31.7.0(", 10))
        {
          data.instantaneousCurrentL1 = strtod(ptrBracket + 1, NULL);
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "1-0:51.7.0(", 10))
        {
          data.instantaneousCurrentL2 = strtod(ptrBracket + 1, NULL);
          data.isDirty = true;
        }
        else if (!strncmp(receivedChars, "1-0:71.7.0(", 10))
        {
          data.instantaneousCurrentL3 = strtod(ptrBracket + 1, NULL);
          data.isDirty = true;
        }
        else if (receivedChars[0] == '!')
        {
          // MISSING: check CRC
          // this prototype assumes OK
          data.isDirty = false;
          if (cbOnNewData) cbOnNewData();
        }
        newData = false;
      }
    }
};
